@@ -1,5 +1,6 @@
 /*
- * Copyright (C) 2020 Alexander Mikhaylenko <alexm@gnome.org>
+ * Copyright (C) 2020 Alice Mikhaylenko <alicem@gnome.org>
+ * Copyright (C) 2023 Purism SPC
  *
  * SPDX-License-Identifier: LGPL-2.1-or-later
  */
@@ -7,7 +8,9 @@
 #include "config.h"
 
 #include "adw-application-window.h"
-#include "adw-window-mixin-private.h"
+
+#include "adw-breakpoint-bin-private.h"
+#include "adw-gizmo-private.h"
 
 /**
  * AdwApplicationWindow:
@@ -24,15 +27,32 @@
  *
  * See [class@Window] for details.
  *
+ * Example of an `AdwApplicationWindow` UI definition:
+ *
+ * ```xml
+ * <object class="AdwApplicationWindow">
+ *   <property name="content">
+ *     <object class="AdwToolbarView">
+ *       <child type="top">
+ *         <object class="AdwHeaderBar"/>
+ *       </child>
+ *       <property name="content">
+ *         <!-- ... -->
+ *       </property>
+ *     </object>
+ *   </property>
+ * </object>
+ * ```
+ *
  * Using [property@Gtk.Application:menubar] is not supported and may result in
  * visual glitches.
- *
- * Since: 1.0
  */
 
 typedef struct
 {
-  AdwWindowMixin *mixin;
+  GtkWidget *titlebar;
+  GtkWidget *bin;
+  GtkWidget *content;
 } AdwApplicationWindowPrivate;
 
 static void adw_application_window_buildable_init (GtkBuildableIface *iface);
@@ -46,12 +66,17 @@ static GtkBuildableIface *parent_buildable_iface;
 enum {
   PROP_0,
   PROP_CONTENT,
+  PROP_CURRENT_BREAKPOINT,
   LAST_PROP,
 };
 
 static GParamSpec *props[LAST_PROP];
 
-#define ADW_GET_WINDOW_MIXIN(obj) (((AdwApplicationWindowPrivate *) adw_application_window_get_instance_private (ADW_APPLICATION_WINDOW (obj)))->mixin)
+static void
+notify_current_breakpoint_cb (AdwApplicationWindow *self)
+{
+  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_CURRENT_BREAKPOINT]);
+}
 
 static void
 adw_application_window_size_allocate (GtkWidget *widget,
@@ -59,21 +84,20 @@ adw_application_window_size_allocate (GtkWidget *widget,
                                       int        height,
                                       int        baseline)
 {
-  adw_window_mixin_size_allocate (ADW_GET_WINDOW_MIXIN (widget),
-                                  width,
-                                  height,
-                                  baseline);
-}
-
-static void
-adw_application_window_finalize (GObject *object)
-{
-  AdwApplicationWindow *self = (AdwApplicationWindow *)object;
+  AdwApplicationWindow *self = ADW_APPLICATION_WINDOW (widget);
   AdwApplicationWindowPrivate *priv = adw_application_window_get_instance_private (self);
 
-  g_clear_object (&priv->mixin);
+  /* We don't want to allow any other titlebar */
+  if (gtk_window_get_titlebar (GTK_WINDOW (self)) != priv->titlebar)
+    g_error ("gtk_window_set_titlebar() is not supported for AdwApplicationWindow");
 
-  G_OBJECT_CLASS (adw_application_window_parent_class)->finalize (object);
+  if (gtk_window_get_child (GTK_WINDOW (self)) != priv->bin)
+    g_error ("gtk_window_set_child() is not supported for AdwApplicationWindow");
+
+  GTK_WIDGET_CLASS (adw_application_window_parent_class)->size_allocate (widget,
+                                                                         width,
+                                                                         height,
+                                                                         baseline);
 }
 
 static void
@@ -88,6 +112,9 @@ adw_application_window_get_property (GObject    *object,
   case PROP_CONTENT:
     g_value_set_object (value, adw_application_window_get_content (self));
     break;
+  case PROP_CURRENT_BREAKPOINT:
+    g_value_set_object (value, adw_application_window_get_current_breakpoint (self));
+    break;
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
@@ -116,7 +143,6 @@ adw_application_window_class_init (AdwApplicationWindowClass *klass)
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
   GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
 
-  object_class->finalize = adw_application_window_finalize;
   object_class->get_property = adw_application_window_get_property;
   object_class->set_property = adw_application_window_set_property;
   widget_class->size_allocate = adw_application_window_size_allocate;
@@ -126,14 +152,24 @@ adw_application_window_class_init (AdwApplicationWindowClass *klass)
    *
    * The content widget.
    *
-   * Since: 1.0
+   * This property should always be used instead of [property@Gtk.Window:child].
    */
   props[PROP_CONTENT] =
-    g_param_spec_object ("content",
-                         "Content",
-                         "The content widget",
+    g_param_spec_object ("content", NULL, NULL,
                          GTK_TYPE_WIDGET,
-                         G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY);
+                         G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
+
+  /**
+   * AdwApplicationWindow:current-breakpoint: (attributes org.gtk.Property.get=adw_application_window_get_current_breakpoint)
+   *
+   * The current breakpoint.
+   *
+   * Since: 1.4
+   */
+  props[PROP_CURRENT_BREAKPOINT] =
+    g_param_spec_object ("current-breakpoint", NULL, NULL,
+                         ADW_TYPE_BREAKPOINT,
+                         G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
 
   g_object_class_install_properties (object_class, LAST_PROP, props);}
 
@@ -142,8 +178,18 @@ adw_application_window_init (AdwApplicationWindow *self)
 {
   AdwApplicationWindowPrivate *priv = adw_application_window_get_instance_private (self);
 
-  priv->mixin = adw_window_mixin_new (GTK_WINDOW (self),
-                                      GTK_WINDOW_CLASS (adw_application_window_parent_class));
+  priv->titlebar = adw_gizmo_new_with_role ("nothing", GTK_ACCESSIBLE_ROLE_PRESENTATION,
+                                            NULL, NULL, NULL, NULL, NULL, NULL);
+  gtk_widget_set_visible (priv->titlebar, FALSE);
+  gtk_window_set_titlebar (GTK_WINDOW (self), priv->titlebar);
+
+  priv->bin = adw_breakpoint_bin_new ();
+  adw_breakpoint_bin_set_warning_widget (ADW_BREAKPOINT_BIN (priv->bin),
+                                         GTK_WIDGET (self));
+  gtk_window_set_child (GTK_WINDOW (self), priv->bin);
+
+  g_signal_connect_swapped (priv->bin, "notify::current-breakpoint",
+                            G_CALLBACK (notify_current_breakpoint_cb), self);
 
   gtk_application_window_set_show_menubar (GTK_APPLICATION_WINDOW (self), FALSE);
 }
@@ -158,6 +204,9 @@ adw_application_window_buildable_add_child (GtkBuildable *buildable,
     GTK_BUILDER_WARN_INVALID_CHILD_TYPE (buildable, type);
   else if (GTK_IS_WIDGET (child))
     adw_application_window_set_content (ADW_APPLICATION_WINDOW (buildable), GTK_WIDGET (child));
+  else if (ADW_IS_BREAKPOINT (child))
+    adw_application_window_add_breakpoint (ADW_APPLICATION_WINDOW (buildable),
+                                           g_object_ref (ADW_BREAKPOINT (child)));
   else
     parent_buildable_iface->add_child (buildable, builder, child, type);
 }
@@ -177,8 +226,6 @@ adw_application_window_buildable_init (GtkBuildableIface *iface)
  * Creates a new `AdwApplicationWindow` for @app.
  *
  * Returns: the newly created `AdwApplicationWindow`
- *
- * Since: 1.0
  */
 GtkWidget *
 adw_application_window_new (GtkApplication *app)
@@ -196,17 +243,25 @@ adw_application_window_new (GtkApplication *app)
  * Sets the content widget of @self.
  *
  * This method should always be used instead of [method@Gtk.Window.set_child].
- *
- * Since: 1.0
  */
 void
 adw_application_window_set_content (AdwApplicationWindow *self,
                                     GtkWidget            *content)
 {
+  AdwApplicationWindowPrivate *priv;
+
   g_return_if_fail (ADW_IS_APPLICATION_WINDOW (self));
   g_return_if_fail (content == NULL || GTK_IS_WIDGET (content));
 
-  adw_window_mixin_set_content (ADW_GET_WINDOW_MIXIN (self), content);
+  if (content)
+    g_return_if_fail (gtk_widget_get_parent (content) == NULL);
+
+  priv = adw_application_window_get_instance_private (self);
+
+  if (adw_application_window_get_content (self) == content)
+    return;
+
+  adw_breakpoint_bin_set_child (ADW_BREAKPOINT_BIN (priv->bin), content);
 
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_CONTENT]);
 }
@@ -220,13 +275,60 @@ adw_application_window_set_content (AdwApplicationWindow *self,
  * This method should always be used instead of [method@Gtk.Window.get_child].
  *
  * Returns: (nullable) (transfer none): the content widget of @self
- *
- * Since: 1.0
  */
 GtkWidget *
 adw_application_window_get_content (AdwApplicationWindow *self)
 {
+  AdwApplicationWindowPrivate *priv;
+
   g_return_val_if_fail (ADW_IS_APPLICATION_WINDOW (self), NULL);
 
-  return adw_window_mixin_get_content (ADW_GET_WINDOW_MIXIN (self));
+  priv = adw_application_window_get_instance_private (self);
+
+  return adw_breakpoint_bin_get_child (ADW_BREAKPOINT_BIN (priv->bin));
+}
+
+/**
+ * adw_application_window_add_breakpoint:
+ * @self: an application window
+ * @breakpoint: (transfer full): the breakpoint to add
+ *
+ * Adds @breakpoint to @self.
+ *
+ * Since: 1.4
+ */
+void
+adw_application_window_add_breakpoint (AdwApplicationWindow *self,
+                                       AdwBreakpoint        *breakpoint)
+{
+  AdwApplicationWindowPrivate *priv;
+
+  g_return_if_fail (ADW_IS_APPLICATION_WINDOW (self));
+  g_return_if_fail (ADW_IS_BREAKPOINT (breakpoint));
+
+  priv = adw_application_window_get_instance_private (self);
+
+  adw_breakpoint_bin_add_breakpoint (ADW_BREAKPOINT_BIN (priv->bin), breakpoint);
+}
+
+/**
+ * adw_application_window_get_current_breakpoint: (attributes org.gtk.Method.get_property=current-breakpoint)
+ * @self: an application window
+ *
+ * Gets the current breakpoint.
+ *
+ * Returns: (nullable) (transfer none): the current breakpoint
+ *
+ * Since: 1.4
+ */
+AdwBreakpoint *
+adw_application_window_get_current_breakpoint (AdwApplicationWindow *self)
+{
+  AdwApplicationWindowPrivate *priv;
+
+  g_return_val_if_fail (ADW_IS_APPLICATION_WINDOW (self), NULL);
+
+  priv = adw_application_window_get_instance_private (self);
+
+  return adw_breakpoint_bin_get_current_breakpoint (ADW_BREAKPOINT_BIN (priv->bin));
 }
