@@ -11,15 +11,18 @@
 
 #include "adw-animation-util.h"
 #include "adw-action-row.h"
+#include "adw-breakpoint-bin.h"
 #include "adw-leaflet.h"
+#include "adw-navigation-view.h"
 #include "adw-preferences-group-private.h"
 #include "adw-preferences-page-private.h"
 #include "adw-toast-overlay.h"
-#include "adw-view-switcher.h"
-#include "adw-view-switcher-bar.h"
-#include "adw-view-switcher-title.h"
+#include "adw-view-stack.h"
 #include "adw-widget-utils-private.h"
 
+#define VIEW_SWITCHER_PAGE_THRESHOLD 110
+#define VIEW_SWITCHER_FALLBACK_THRESHOLD 400
+
 /**
  * AdwPreferencesWindow:
  *
@@ -42,8 +45,8 @@
 typedef struct
 {
   AdwToastOverlay *toast_overlay;
-  AdwLeaflet *subpages_leaflet;
-  GtkWidget *preferences;
+  AdwNavigationView *subpages_nav_view;
+  GtkWidget *breakpoint_bin;
   GtkStack *content_stack;
   AdwViewStack *pages_stack;
   GtkToggleButton *search_button;
@@ -51,8 +54,10 @@ typedef struct
   GtkListBox *search_results;
   GtkStack *search_stack;
   GtkStack *title_stack;
-  AdwViewSwitcherBar *view_switcher_bar;
-  AdwViewSwitcherTitle *view_switcher_title;
+  GtkWidget *view_switcher_stack;
+  GtkWidget *view_switcher;
+  GtkWidget *title;
+  AdwBreakpoint *breakpoint;
 
   gboolean search_enabled;
   gboolean can_navigate_back;
@@ -60,6 +65,9 @@ typedef struct
   GtkFilter *filter;
   GtkFilterListModel *filter_model;
 
+  int n_pages;
+
+  AdwLeaflet *subpages_leaflet;
   GtkWidget *subpage;
 } AdwPreferencesWindowPrivate;
 
@@ -205,15 +213,15 @@ get_n_pages (AdwPreferencesWindow *self)
   return count;
 }
 
-static gchar *
+static char *
 create_search_row_subtitle (AdwPreferencesWindow *self,
-                            GtkWidget            *row)
+                            AdwPreferencesRow    *row)
 {
   GtkWidget *group, *page;
   const char *group_title = NULL;
   char *page_title = NULL;
 
-  group = gtk_widget_get_ancestor (row, ADW_TYPE_PREFERENCES_GROUP);
+  group = gtk_widget_get_ancestor (GTK_WIDGET (row), ADW_TYPE_PREFERENCES_GROUP);
 
   if (group) {
     group_title = adw_preferences_group_get_title (ADW_PREFERENCES_GROUP (group));
@@ -232,6 +240,14 @@ create_search_row_subtitle (AdwPreferencesWindow *self,
     else
       page_title = g_strdup (title);
 
+    if (adw_preferences_row_get_use_markup (row)) {
+      char *tmp = page_title;
+
+      page_title = g_markup_escape_text (page_title, -1);
+
+      g_free (tmp);
+    }
+
     if (g_strcmp0 (page_title, "") == 0)
       g_clear_pointer (&page_title, g_free);
   }
@@ -262,13 +278,17 @@ new_search_row_for_preference (AdwPreferencesRow    *row,
 
   g_assert (ADW_IS_PREFERENCES_ROW (row));
 
-  subtitle = create_search_row_subtitle (self, GTK_WIDGET (row));
+  subtitle = create_search_row_subtitle (self, row);
   page = gtk_widget_get_ancestor (GTK_WIDGET (row), ADW_TYPE_PREFERENCES_PAGE);
 
   widget = ADW_ACTION_ROW (adw_action_row_new ());
   gtk_list_box_row_set_activatable (GTK_LIST_BOX_ROW (widget), TRUE);
-  g_object_bind_property (row, "title", widget, "title", G_BINDING_SYNC_CREATE);
-  g_object_bind_property (row, "use-underline", widget, "use-underline", G_BINDING_SYNC_CREATE);
+  adw_preferences_row_set_use_markup (ADW_PREFERENCES_ROW (widget),
+                                      adw_preferences_row_get_use_markup (row));
+  adw_preferences_row_set_use_underline (ADW_PREFERENCES_ROW (widget),
+                                         adw_preferences_row_get_use_underline (row));
+  adw_preferences_row_set_title (ADW_PREFERENCES_ROW (widget),
+                                 adw_preferences_row_get_title (row));
   adw_action_row_set_subtitle (widget, subtitle);
 
   g_object_set_data (G_OBJECT (widget), "page", page);
@@ -321,7 +341,7 @@ search_results_unmap (AdwPreferencesWindow *self)
 }
 
 static void
-try_remove_subpages (AdwPreferencesWindow *self)
+try_remove_legacy_subpages (AdwPreferencesWindow *self)
 {
   AdwPreferencesWindowPrivate *priv = adw_preferences_window_get_instance_private (self);
   GtkWidget *child;
@@ -329,7 +349,7 @@ try_remove_subpages (AdwPreferencesWindow *self)
   if (adw_leaflet_get_child_transition_running (priv->subpages_leaflet))
     return;
 
-  if (adw_leaflet_get_visible_child (priv->subpages_leaflet) == priv->preferences)
+  if (adw_leaflet_get_visible_child (priv->subpages_leaflet) == GTK_WIDGET (priv->subpages_nav_view))
     priv->subpage = NULL;
 
   child = gtk_widget_get_first_child (GTK_WIDGET (priv->subpages_leaflet));
@@ -338,21 +358,39 @@ try_remove_subpages (AdwPreferencesWindow *self)
 
     child = gtk_widget_get_next_sibling (child);
 
-    if (page != priv->preferences && page != priv->subpage)
+    if (page != GTK_WIDGET (priv->subpages_nav_view) && page != priv->subpage)
       adw_leaflet_remove (priv->subpages_leaflet, page);
   }
 }
 
 static void
-subpages_leaflet_child_transition_running_cb (AdwPreferencesWindow *self)
+update_view_switcher (AdwPreferencesWindow *self)
 {
-  try_remove_subpages (self);
-}
+  AdwPreferencesWindowPrivate *priv = adw_preferences_window_get_instance_private (self);
+  AdwBreakpoint *breakpoint;
+  AdwBreakpointCondition *main_condition, *fallback_condition, *condition;
 
-static void
-subpages_leaflet_visible_child_cb (AdwPreferencesWindow *self)
-{
-  try_remove_subpages (self);
+  main_condition =
+    adw_breakpoint_condition_new_length (ADW_BREAKPOINT_CONDITION_MAX_WIDTH,
+                                         VIEW_SWITCHER_PAGE_THRESHOLD * MAX (1, priv->n_pages),
+                                         ADW_LENGTH_UNIT_PT);
+  fallback_condition =
+    adw_breakpoint_condition_new_length (ADW_BREAKPOINT_CONDITION_MAX_WIDTH,
+                                         VIEW_SWITCHER_FALLBACK_THRESHOLD,
+                                         ADW_LENGTH_UNIT_PX);
+
+  condition = adw_breakpoint_condition_new_or (main_condition, fallback_condition);
+
+  adw_breakpoint_set_condition (priv->breakpoint, condition);
+
+  breakpoint = adw_breakpoint_bin_get_current_breakpoint (ADW_BREAKPOINT_BIN (priv->breakpoint_bin));
+
+  if (!breakpoint && priv->n_pages > 1)
+    gtk_stack_set_visible_child (GTK_STACK (priv->view_switcher_stack), priv->view_switcher);
+  else
+    gtk_stack_set_visible_child (GTK_STACK (priv->view_switcher_stack), priv->title);
+
+  adw_breakpoint_condition_free (condition);
 }
 
 static void
@@ -361,7 +399,7 @@ title_stack_notify_transition_running_cb (AdwPreferencesWindow *self)
   AdwPreferencesWindowPrivate *priv = adw_preferences_window_get_instance_private (self);
 
   if (gtk_stack_get_transition_running (priv->title_stack) ||
-      gtk_stack_get_visible_child (priv->title_stack) != GTK_WIDGET (priv->view_switcher_title))
+      gtk_stack_get_visible_child (priv->title_stack) != priv->view_switcher_stack)
     return;
 
   gtk_editable_set_text (GTK_EDITABLE (priv->search_entry), "");
@@ -373,7 +411,7 @@ title_stack_notify_visible_child_cb (AdwPreferencesWindow *self)
   AdwPreferencesWindowPrivate *priv = adw_preferences_window_get_instance_private (self);
 
   if (adw_get_enable_animations (GTK_WIDGET (priv->title_stack)) ||
-      gtk_stack_get_visible_child (priv->title_stack) != GTK_WIDGET (priv->view_switcher_title))
+      gtk_stack_get_visible_child (priv->title_stack) != priv->view_switcher_stack)
     return;
 
   gtk_editable_set_text (GTK_EDITABLE (priv->search_entry), "");
@@ -597,11 +635,16 @@ adw_preferences_window_class_init (AdwPreferencesWindowClass *klass)
    * <kbd>Alt</kbd>+<kbd>←</kbd> shortcut.
    *
    * For right-to-left locales, gestures and shortcuts are reversed.
+   *
+   * Deprecated: 1.4: Use [property@NavigationPage:can-pop] instead.
+   *
+   * Has no effect for subpages added with
+   * [method@PreferencesWindow.push_subpage].
    */
   props[PROP_CAN_NAVIGATE_BACK] =
     g_param_spec_boolean ("can-navigate-back", NULL, NULL,
                           FALSE,
-                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
+                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_DEPRECATED);
 
   g_object_class_install_properties (object_class, LAST_PROP, props);
 
@@ -612,7 +655,8 @@ adw_preferences_window_class_init (AdwPreferencesWindowClass *klass)
                                                "/org/gnome/Adwaita/ui/adw-preferences-window.ui");
   gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, toast_overlay);
   gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, subpages_leaflet);
-  gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, preferences);
+  gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, subpages_nav_view);
+  gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, breakpoint_bin);
   gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, content_stack);
   gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, pages_stack);
   gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, search_button);
@@ -620,10 +664,12 @@ adw_preferences_window_class_init (AdwPreferencesWindowClass *klass)
   gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, search_results);
   gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, search_stack);
   gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, title_stack);
-  gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, view_switcher_bar);
-  gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, view_switcher_title);
-  gtk_widget_class_bind_template_callback (widget_class, subpages_leaflet_child_transition_running_cb);
-  gtk_widget_class_bind_template_callback (widget_class, subpages_leaflet_visible_child_cb);
+  gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, view_switcher_stack);
+  gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, view_switcher);
+  gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, title);
+  gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, breakpoint);
+  gtk_widget_class_bind_template_callback (widget_class, try_remove_legacy_subpages);
+  gtk_widget_class_bind_template_callback (widget_class, update_view_switcher);
   gtk_widget_class_bind_template_callback (widget_class, title_stack_notify_transition_running_cb);
   gtk_widget_class_bind_template_callback (widget_class, title_stack_notify_visible_child_cb);
   gtk_widget_class_bind_template_callback (widget_class, notify_visible_page_cb);
@@ -735,6 +781,9 @@ adw_preferences_window_add (AdwPreferencesWindow *self,
   g_object_bind_property (page, "title", stack_page, "title", G_BINDING_SYNC_CREATE);
   g_object_bind_property (page, "use-underline", stack_page, "use-underline", G_BINDING_SYNC_CREATE);
   g_object_bind_property (page, "name", stack_page, "name", G_BINDING_SYNC_CREATE);
+
+  priv->n_pages++;
+  update_view_switcher (self);
 }
 
 /**
@@ -759,6 +808,9 @@ adw_preferences_window_remove (AdwPreferencesWindow *self,
     adw_view_stack_remove (priv->pages_stack, GTK_WIDGET (page));
   else
     ADW_CRITICAL_CANNOT_REMOVE_CHILD (self, page);
+
+  priv->n_pages--;
+  update_view_switcher (self);
 }
 
 /**
@@ -915,6 +967,10 @@ adw_preferences_window_set_search_enabled (AdwPreferencesWindow *self,
  * <kbd>Alt</kbd>+<kbd>←</kbd> shortcut.
  *
  * For right-to-left locales, gestures and shortcuts are reversed.
+ *
+ * Deprecated: 1.4: Use [method@NavigationPage.set_can_pop] instead.
+ *
+ * Has no effect for subpages added with [method@PreferencesWindow.push_subpage].
  */
 void
 adw_preferences_window_set_can_navigate_back (AdwPreferencesWindow *self,
@@ -943,6 +999,8 @@ adw_preferences_window_set_can_navigate_back (AdwPreferencesWindow *self,
  * Gets whether gestures and shortcuts for closing subpages are enabled.
  *
  * Returns: whether gestures and shortcuts are enabled.
+ *
+ * Deprecated: 1.4: Use [method@NavigationPage.get_can_pop] instead.
  */
 gboolean
 adw_preferences_window_get_can_navigate_back (AdwPreferencesWindow *self)
@@ -965,6 +1023,8 @@ adw_preferences_window_get_can_navigate_back (AdwPreferencesWindow *self)
  *
  * The transition can be cancelled by the user, in which case visible child will
  * change back to the previously visible child.
+ *
+ * Deprecated: 1.4: Use [method@PreferencesWindow.push_subpage] instead.
  */
 void
 adw_preferences_window_present_subpage (AdwPreferencesWindow *self,
@@ -998,6 +1058,8 @@ adw_preferences_window_present_subpage (AdwPreferencesWindow *self,
  * Closes the current subpage.
  *
  * If there is no presented subpage, this does nothing.
+ *
+ * Deprecated: 1.4: Use [method@PreferencesWindow.pop_subpage] instead.
  */
 void
 adw_preferences_window_close_subpage (AdwPreferencesWindow *self)
@@ -1011,7 +1073,54 @@ adw_preferences_window_close_subpage (AdwPreferencesWindow *self)
   if (priv->subpage == NULL)
     return;
 
-  adw_leaflet_set_visible_child (priv->subpages_leaflet, priv->preferences);
+  adw_leaflet_set_visible_child (priv->subpages_leaflet, GTK_WIDGET (priv->subpages_nav_view));
+}
+
+/**
+ * adw_preferences_window_push_subpage:
+ * @self: a preferences window
+ * @page: the subpage
+ *
+ * Pushes @page onto the subpage stack of @self.
+ *
+ * The page will be automatically removed when popped.
+ *
+ * Since: 1.4
+ */
+void
+adw_preferences_window_push_subpage (AdwPreferencesWindow *self,
+                                     AdwNavigationPage    *page)
+{
+  AdwPreferencesWindowPrivate *priv;
+
+  g_return_if_fail (ADW_IS_PREFERENCES_WINDOW (self));
+  g_return_if_fail (ADW_IS_NAVIGATION_PAGE (page));
+
+  priv = adw_preferences_window_get_instance_private (self);
+
+  adw_navigation_view_push (priv->subpages_nav_view, page);
+}
+
+/**
+ * adw_preferences_window_pop_subpage:
+ * @self: a preferences window
+ *
+ * Pop the visible page from the subpage stack of @self.
+ *
+ * Returns: `TRUE` if a page has been popped
+ *
+ * Since: 1.4
+ */
+gboolean
+adw_preferences_window_pop_subpage (AdwPreferencesWindow *self)
+{
+  AdwPreferencesWindowPrivate *priv;
+
+  g_return_val_if_fail (ADW_IS_PREFERENCES_WINDOW (self), FALSE);
+
+  priv = adw_preferences_window_get_instance_private (self);
+
+  return adw_navigation_view_pop (priv->subpages_nav_view);
 }
 
 /**
@@ -1035,4 +1144,3 @@ adw_preferences_window_add_toast (AdwPreferencesWindow *self,
 
   adw_toast_overlay_add_toast (priv->toast_overlay, toast);
 }
-
